# GloFlow application and media management/publishing platform
# Copyright (C) 2020 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

import tensorflow as tf
from tensorflow.keras import datasets

import sklearn.datasets
import sklearn.model_selection

#---------------------------
# GF_IMAGES_JOBS

print('''
RUN WITH - 'LD_LIBRARY_PATH=./../../../rust/build python3 gf_simple_model.py'
''')

# os.environ["LD_LIBRARY_PATH"] = "%s/../../../rust/build"%(modd_str)
# assert "LD_LIBRARY_PATH" in os.environ.keys()

# FIX!! - this is super temporary!! as soon as possible fix this, where the 
#         PY extension lib file is not copied explicitly to the current directory.
#         see why using LD_LIBRARY_PATH doesnt work for gf_images_jobs_py.so, 
#         but does work for libtensorflow.so
import delegator
gf_py_libs_path_str = "%s/../../../rust/build"%(modd_str)
delegator.run("cp %s/gf_images_jobs_py.so %s"%(gf_py_libs_path_str, modd_str))

print("loading gf_images_jobs_py.so")
import gf_images_jobs_py as gf_images_jobs

#---------------------------

#----------------------------------------------
# LOAD_GENERATED
def load__generated(p_generate_bool = False):

    print("load generated...")

    dataset_target_dir_path_str = "%s/test/data/output/generated"%(modd_str)
    dataset_train_file_path_str = "%s/tfrecords/test__train.tfrecords"%(dataset_target_dir_path_str)

    # GENERATE
    if p_generate_bool:
        dataset_name_str = "test"
        classes_lst      = [
            "rect",
            "circle"
        ]

        elements_num_int = 100
        img_width_int    = 32
        img_height_int   = 32




        # GENERATE_ML_DATASET
        gf_images_jobs.generate_ml_dataset(dataset_name_str,
            classes_lst,
            elements_num_int,
            img_width_int,
            img_height_int,
            dataset_target_dir_path_str)


    print("loading dataset...")
    print(dataset_train_file_path_str)


    assert os.path.isfile(dataset_train_file_path_str)
    dataset = tf.data.TFRecordDataset([dataset_train_file_path_str],

        # scalar representing the number of files to read in parallel.
        # If greater than one, the records of files read in parallel
        # are outputted in an interleaved order
		num_parallel_reads = 32)

    print(dataset)
    return dataset

#----------------------------------------------
# LOAD__CIFAR10
def load__cifar10():

    (train_images, train_labels), (test_images, test_labels) = datasets.cifar10.load_data()

    # Normalize pixel values to be between 0 and 1
    train_images, test_images = train_images / 255.0, test_images / 255.0


    #----------------------------------------------
    def show_dataset():
        class_names = [
            "airplane",
            "automobile",
            "bird",
            "cat",
            "deer",
            "dog",
            "frog",
            "horse",
            "ship",
            "truck"
        ]

        plt.figure(figsize=(10, 10))
        for i in range(25):

            plt.subplot(5,5,i+1)
            plt.xticks([])
            plt.yticks([])
            plt.grid(False)

            plt.imshow(train_images[i], cmap = plt.cm.binary)

            # The CIFAR labels happen to be arrays, which is why you need the extra index
            plt.xlabel(class_names[train_labels[i][0]])

        plt.show()

    #----------------------------------------------
    # show_dataset()

    data_map = {
        "train_images": train_images,
        "train_labels": train_labels,
        "test_images":  test_images,
        "test_labels":  test_labels
    }

    return data_map

#----------------------------------------------
# LOAD__MNIST
def load__mnist():

    # https://scikit-learn.org/stable/modules/generated/sklearn.datasets.load_digits.html
    dataset = sklearn.datasets.load_digits()
    imgs_examples_ndarr = dataset.data
    imgs_labels_ndarr   = dataset.target




    # https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html
    # Split arrays or matrices into random train and test subsets.
    # shuffle - randomize data before splitting it
    imgs_train, imgs_test, labels_train, labels_test = sklearn.model_selection.train_test_split(imgs_examples_ndarr,
        imgs_labels_ndarr,
        shuffle=True)


    return imgs_train, imgs_test, labels_train, labels_test


# GloFlow application and media management/publishing platform
# Copyright (C) 2024 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os

from google.oauth2 import service_account
from google.cloud.aiplatform.gapic.schema import predict
from google.cloud import vision
# from google.cloud import aiplatform

#--------------------------------------------
# DESCRIBE_IMAGE

def describe_image(p_url_str, p_ai_client):
    
    print(f"describing image {p_url_str} ...")

    image = vision.Image()
    image.source.image_uri = p_url_str

    response = p_ai_client.label_detection(image=image)
    labels = response.label_annotations

    print("Labels:")
    labels_lst = []
    for label in labels:
        
        print(">+>>>>>>>>>", label.score, label.description)
        
        label_str = label.description
        labels_lst.append((label_str, label.score))
    return labels_lst
        
#--------------------------------------------
# CLASSIFY_IMAGE
# not used currently

def classify_images(p_image_path_local_str,
    p_project_id_str,
    p_model_id_str,
    p_client):
    assert os.path.isfile(p_image_path_local_str)

    # MODEL
    model_name = p_client.model_path(p_project_id_str, "us-central1", p_model_id_str)

    with open(p_image_path_local_str, "rb") as f:
        file_content = f.read()

    request = predict.instance.ImageClassificationPredictionInstance(
        content=file_content
    ).to_value()

    instances = [
        request
    ]

    # PREDICT
    response = p_client.predict(endpoint=model_name, instances=instances)

    # process responses
    for prediction in response.predictions:
        print("Prediction result:", prediction)

#--------------------------------------------
def init(p_key_info_map):
    
    credentials  = service_account.Credentials.from_service_account_info(p_key_info_map)
    
    # client = bigquery.Client(credentials=credentials, project=credentials.project_id)
    # client = aiplatform.gapic.PredictionServiceClient(credentials=credentials)
    client = vision.ImageAnnotatorClient(credentials=credentials)
    
    return client


# GloFlow application and media management/publishing platform
# Copyright (C) 2020 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

import numpy as np
import skimage.color
import skimage.io
import requests

#----------------------------------------------
def download_test_img():

    target_test_img_str = "%s/../data/input/1234cd19517b939d3eb726c817985fe4_thumb_medium.jpeg"%(modd_str)
    url_str             = "http://gf--img.s3-website-us-east-1.amazonaws.com/thumbnails/1234cd19517b939d3eb726c817985fe4_thumb_medium.jpeg"
    
    if not os.path.isfile(target_test_img_str):
        
        r = requests.get(url_str)
        assert r.status_code == 200
        f = open(target_test_img_str, "wb")
        f.write(r.content)
        f.close()
    
    # LOAD_FROM_FILE
    input_img = skimage.io.imread(target_test_img_str)
    assert isinstance(input_img, np.ndarray)
    print("input img - %s"%(str(input_img.shape)))


    img_grayscale = skimage.color.rgb2gray(input_img)
    img_croped    = img_crop(img_grayscale, 8, 8)

    
    # add another dimension to data, first one, to have a batch-like np array
    input_img_batch = np.expand_dims(img_croped, 0)

    return input_img_batch, target_test_img_str

#----------------------------------------------
def load_data():

    input_img_path_str = "%s/../../../../../rust/gf_images_jobs/test/data/input/test_img_1.jpeg"%(modd_str)
    assert os.path.isfile(input_img_path_str)

    
    # LOAD_FROM_FILE
    input_img = skimage.io.imread(input_img_path_str)
    assert isinstance(input_img, np.ndarray)
    print("input img - %s"%(str(input_img.shape)))

   
    # CROP
    img_croped            = img_crop(input_img, 64, 64)
    input_img_transformed = img_croped.astype("float32")

    # plt.imshow(input_img_transformed.astype("uint8"))
    # plt.show()

    input_img_batch = np.expand_dims(input_img_transformed, 0)
    print(input_img_batch.shape)

    return input_img_batch
    
#----------------------------------------------
def img_crop(p_img_np, p_width_int, p_height_int):
    h = p_img_np.shape[0]
    w = p_img_np.shape[1]

    crop_origin_x = w//2 - p_width_int//2
    crop_origin_y = h//2 - p_height_int//2
    img_crop = p_img_np[crop_origin_x:crop_origin_x+p_width_int, 
        crop_origin_y:crop_origin_y+p_height_int]
        
    return img_crop


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

import subprocess
from colored import fg, bg, attr

sys.path.append("%s/../../gf_core"%(modd_str))
import gf_core_cli

import gf_os_docker

#-------------------------------------------------------------
# PULL

def pull(p_image__full_name_str,
	p_log_fun,
	p_docker_user_str   = None,
	p_docker_pass_str   = None,
	p_exit_on_fail_bool = True,
	p_docker_sudo_bool  = False):

	# often times public containers are being pulled, so no login is needed for that and 
	# callers dont submit their credentials.
	if not p_docker_pass_str == None and not p_docker_pass_str == "":
		gf_os_docker.login(p_docker_user_str,
			p_docker_pass_str,
			p_exit_on_fail_bool = True,
			p_docker_sudo_bool  = p_docker_sudo_bool)

	# DOCKER_PULL
	cmd_lst = []
	if p_docker_sudo_bool:
		cmd_lst.append("sudo")
	
	cmd_lst.extend([
		"docker pull",
		p_image__full_name_str
	])
	c_pull = " ".join(cmd_lst)
	p_log_fun("INFO", "cmd - %s"%(c_pull))

	p = subprocess.Popen(c_pull,
		shell   = True,
		stdout  = subprocess.PIPE,
		bufsize = 1)

	for line in p.stdout:
		clean_line_str = line.strip()
		print(clean_line_str)

	if p_exit_on_fail_bool:
		if not p.returncode == None and not p.returncode == 0:
			exit()

#-------------------------------------------------------------
# BUILD

def build(p_app_name_str,
	p_app_build_meta_map,
	p_log_fun,
	p_app_web_meta_map    = None,
	p_user_name_str       = "local",
	p_git_commit_hash_str = None,
	p_exit_on_fail_bool   = False,
	p_docker_sudo_bool    = False):
	p_log_fun("INFO", f"p_app_name_str - {p_app_name_str}")
	assert isinstance(p_app_name_str,       str)
	assert isinstance(p_app_build_meta_map, dict)

	#------------------
	# META

	is_service_bool = False
	if "service_name_str" in p_app_build_meta_map:
		assert "service_base_dir_str" in p_app_build_meta_map.keys()
		is_service_bool = True

		
	if is_service_bool:
		service_name_str     = p_app_build_meta_map["service_name_str"]
		service_base_dir_str = p_app_build_meta_map["service_base_dir_str"]
		assert os.path.isdir(service_base_dir_str)

	# service_dockerfile_path_str = "%s/Dockerfile"%(service_base_dir_str)
	service_dockerfile_path_str = get_service_dockerfile(p_app_build_meta_map)
	
	#------------------
	# COPY_FILES_TO_DIR
	if "copy_to_dir_lst" in p_app_build_meta_map.keys():

		copy_to_dir_lst = p_app_build_meta_map["copy_to_dir_lst"]
		
		copy_files(copy_to_dir_lst)

	#------------------
	# PREPARE_WEB_FILES
	if not p_app_web_meta_map == None:
		assert isinstance(p_app_web_meta_map, dict)
		assert "pages_map" in p_app_web_meta_map.keys()
		pages_map = p_app_web_meta_map["pages_map"]

		prepare_web_files(pages_map,
			service_base_dir_str,
			p_log_fun,
			p_docker_sudo_bool = p_docker_sudo_bool)

	#------------------
	# IMAGE_FULL_NAMES
		
	if is_service_bool:
		image_name_str = service_name_str
	else:
		assert "cont_image_name_str" in p_app_build_meta_map.keys()
		image_name_str = p_app_build_meta_map["cont_image_name_str"]
		
	image_full_names_lst = get_image_full_names(image_name_str,
		p_app_build_meta_map,
		p_user_name_str,
		p_git_commit_hash_str = p_git_commit_hash_str)

	#------------------
	# BUILD_ARGS
	build_args_map = {}

	# BASE_IMAGE_TAG - tag of the base image from which the main image thats being
	#                  built from is inheriting from.
	if not p_git_commit_hash_str == None:
		build_args_map["GF_BASE_IMAGE_TAG"] = p_git_commit_hash_str
	else:
		build_args_map["GF_BASE_IMAGE_TAG"] = "latest"
	
	#------------------
	# DOCKER_BUILD
	gf_os_docker.build_image(image_full_names_lst,
		service_dockerfile_path_str,
		p_log_fun,
		p_build_args_map    = build_args_map,
		p_exit_on_fail_bool = p_exit_on_fail_bool,
		p_docker_sudo_bool  = p_docker_sudo_bool)

	#------------------

#-------------------------------------------------------------
# PUBLISH

def publish(p_app_name_str,
	p_app_build_meta_map,
	p_docker_user_str,
	p_docker_pass_str,
	p_log_fun,
	p_git_commit_hash_str = None,
	p_exit_on_fail_bool   = False,
	p_docker_sudo_bool    = False):
	p_log_fun("INFO", "p_app_name_str - %s"%(p_app_name_str))
	assert isinstance(p_app_build_meta_map, dict)

	if "service_name_str" in p_app_build_meta_map.keys():
		service_name_str = p_app_build_meta_map["service_name_str"]
		image_name_str   = service_name_str
	else:
		image_name_str = p_app_name_str

	# service_version_str = p_app_build_meta_map["version_str"]
	#
	# image_tag_str = None
	# if not p_git_commit_hash_str == None:
	# 	image_tag_str = p_git_commit_hash_str
	# else:
	# 	service_version_str = p_app_build_meta_map["version_str"]
	# 	image_tag_str       = service_version_str

	image_full_names_lst = get_image_full_names(image_name_str,
		p_app_build_meta_map,
		p_docker_user_str,
		p_git_commit_hash_str = p_git_commit_hash_str)
	assert isinstance(image_full_names_lst, list)
	
	for image_full_name_str in image_full_names_lst:
		
		# DOCKER_PUSH
		gf_os_docker.push(image_full_name_str,
			p_docker_user_str,
			p_docker_pass_str,
			p_log_fun,
			p_exit_on_fail_bool = p_exit_on_fail_bool,
			p_docker_sudo_bool  = p_docker_sudo_bool)

#-------------------------------------------------------------
# GET_IMAGE_FULL_NAMES

def get_image_full_names(p_image_name_str,
	p_app_build_meta_map,
	p_user_name_str,
	p_git_commit_hash_str = None):
	assert isinstance(p_image_name_str, str)
	assert isinstance(p_app_build_meta_map, dict)
	assert isinstance(p_user_name_str, str)

	# IMAGE_TAG
	image_tag_str = None
	if not p_git_commit_hash_str == None:
		# if a git commit hash was supplied, tag the image with that
		image_tag_str = p_git_commit_hash_str
	else:
		service_version_str = p_app_build_meta_map["version_str"]

		# assert len(service_version_str.split(".")) == 4 # format x.x.x.x
		image_tag_str = service_version_str

	image_full_names_lst = []

	# standard name
	image_full_name_str = "%s/%s:%s"%(p_user_name_str, p_image_name_str, image_tag_str)
	image_full_names_lst.append(image_full_name_str)

	# IMPORTANT!! - "latest" name - its important to always havea a "latest" image that points
	#               to the most up-to-date container image for use in situations when we dont know
	#               the version number or git commit hash or some other tag.
	if not image_tag_str == "latest":
		image_full_name_latest_str = "%s/%s:latest"%(p_user_name_str, p_image_name_str)
		image_full_names_lst.append(image_full_name_latest_str)

	return image_full_names_lst

#-------------------------------------------------------------
def copy_files(p_copy_to_dir_lst):
	assert isinstance(p_copy_to_dir_lst, list)

	print("")
	print("             COPY FILES")
	for src, target_dir_str in p_copy_to_dir_lst:
		if not os.path.isdir(target_dir_str):
			gf_core_cli.run("mkdir -p %s"%(target_dir_str))

		# COPY_DIR
		if os.path.isdir(src):
			src_dir_str = src
			gf_core_cli.run("cp -r %s %s"%(src_dir_str, target_dir_str))

		# COPY_FILE
		else:
			src_file_str = src
			gf_core_cli.run("cp %s %s"%(src_file_str, target_dir_str))

#-------------------------------------------------------------
# PREPARE_WEB_FILES

def prepare_web_files(p_pages_map,
	p_service_base_dir_str,
	p_log_fun,
	p_docker_sudo_bool = False):
	p_log_fun("FUN_ENTER", "gf_containers.prepare_web_files()")
	assert isinstance(p_pages_map, dict)
	assert os.path.dirname(p_service_base_dir_str)

	for pg_name_str, pg_info_map in p_pages_map.items():
		print(f"======== {fg('green')}{'%s'%(pg_name_str)}{attr(0)}")
		assert isinstance(pg_info_map, dict)
		assert "build_dir_str" in pg_info_map.keys()
		assert os.path.isdir(pg_info_map["build_dir_str"])

		build_dir_str = os.path.abspath(pg_info_map["build_dir_str"])

		#------------------
		# CREATE_TARGET_DIR
		target_dir_str = os.path.abspath(f"{p_service_base_dir_str}/static")
		gf_core_cli.run(f"mkdir -p {target_dir_str}")

		#------------------
		# COPY_PAGE_WEB_CODE
		gf_core_cli.run(f"cp -r {build_dir_str}/* {target_dir_str}")

		#------------------
		
	#------------------
	# MOVE_TEMPLATES_OUT_OF_STATIC

	# IMPORTANT!! - templates should not be in the static/ dir, which would make them servable
	#               over HTTP which we dont want. instead its moved out of the static/ dir 
	#               to its parent dir where its private.
	#               templates are originally in the static/ dir because durring the build process they were
	#               handled together with other static content (html/css/js files) and as output moved
	#               into that static/ dir from other locations while in development.
	gf_core_cli.run("rm -rf %s/../templates"%(target_dir_str)) # remove existing templates build dir
	gf_core_cli.run("mv %s/templates %s/.."%(target_dir_str, target_dir_str))
	
	#------------------

#-------------------------------------------------------------
def get_service_dockerfile(p_app_build_meta_map):

	if "dockerfile_path_str" in p_app_build_meta_map.keys():
		service_dockerfile_path_str = p_app_build_meta_map["dockerfile_path_str"]
	else:

		service_base_dir_str = p_app_build_meta_map["service_base_dir_str"]
		assert os.path.isdir(service_base_dir_str)

		service_dockerfile_path_str = "%s/Dockerfile"%(service_base_dir_str)

	return service_dockerfile_path_str


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

import signal
import subprocess

from colored import fg, bg, attr

#--------------------------------------------------
def run(p_app_name_str,
    p_test_name_str,
    p_app_meta_map,
    p_aws_s3_creds_map,
    p_exit_on_fail_bool         = False,
    p_dynamic_libs_dir_path_str = os.path.abspath("%s/../../../rust/build"%(modd_str))):
    assert isinstance(p_test_name_str,    str)
    assert isinstance(p_app_meta_map,     dict)
    assert isinstance(p_aws_s3_creds_map, dict)

    print("")
    print(" -- test %s%s%s package"%(fg("green"), p_app_name_str, attr(0)))

    if "test_data_to_serve_dir_str" in p_app_meta_map.keys(): use_test_server_bool = True
    else:                                                     use_test_server_bool = False

    # GO_PACKAGE_DIR
    go_package_dir_path_str = p_app_meta_map["go_path_str"]
    assert os.path.isdir(go_package_dir_path_str)
    print("go_package_dir_path_str - %s"%(go_package_dir_path_str))

    #-------------
    # TEST_HTTP_SERVER - used to server assets/images that various Go functions
    #                    that are tested that do fetching of remote resources.
    
    if use_test_server_bool:
        test_data_dir_str = p_app_meta_map["test_data_to_serve_dir_str"]
        assert os.path.isdir(test_data_dir_str)

        print("")
        print("STARTING TEST DATA PYTHON HTTP SERVER ----------------------------")

        # run the python simple server in the dir where the test data is located, so that its served over http
        c_lst = ["cd %s && python -m SimpleHTTPServer 8000"%(test_data_dir_str)]
        print(" ".join(c_lst))

        # IMPORTANT!! - "cd" and py server are run by the shell, which is their parent process, so a session ID is attached
        #               so that its made a group leader. later when the os.killpg() termination signal is sent to that 
        #               group leader (the shell), its child processes will get shutdown as well (py server).
        #               otherwise the py server will be left running after the tests have finished
        server_p = subprocess.Popen(c_lst, stdout=subprocess.PIPE, preexec_fn=os.setsid, shell=True)
        
    #-------------
    cwd_str = os.getcwd()
    os.chdir(go_package_dir_path_str) #change into the target main package dir

    #-------------
    # CMD
    # ADD!! - per app timeout values, so that in gf_meta.py
    #         a test timeout can be specified per app/package.
    cmd_lst = [
        "go test",
        "-timeout 30s",

        # # "-args" - margs subsequent arguments as ones to pass to the tests themselves
        # "-args",
        # "-mongodb_host=%s"%(p_test_mongodb_host_str)
    ]

    # specific test was selected for running, not all tests
    if not p_test_name_str == "all":
        cmd_lst.append("-v") # verbose
        cmd_lst.append(p_test_name_str)

    c = " ".join(cmd_lst)
    print(c)

    #-------------
    # ENV
    e = os.environ.copy()
    
    # RUST - add compiled libraries (*.so) build directory path, so that they can
    #        be loaded at test run time
    print("dynamic libs (LD_LIBRARY_PATH env var) - %s"%(p_dynamic_libs_dir_path_str))
    e["LD_LIBRARY_PATH"] = p_dynamic_libs_dir_path_str
    
    # AWS_CREDS
    e.update(p_aws_s3_creds_map)
    
    #-------------
    # RUN
    p = subprocess.Popen(c.split(" "), stderr=subprocess.PIPE, env=e)
    
    # IMPORTANT!! - stderr is used and read, because thats what contains the log lines from Go programs that has
    #               color codes preserved in log lines.
    for l in iter(p.stderr.readline, ""):
        print(l.rstrip())

    # if not p.stderr == None: print '%sTESTS FAILED%s >>>>>>>\n'%(fg('red'), attr(0))

    p.wait()          # has to be called so that p.returncode is set
    os.chdir(cwd_str) # return to initial dir

    #-------------

    # kill HTTP test server used to serve assets that need to come over HTTP
    if use_test_server_bool:
        os.killpg(server_p.pid, signal.SIGTERM)


    # in certain scenarios (such as CI) we want this test run to fail 
    # completelly in case "go test" returns a non-zero return code (failed test).
    # this way CI pipeline will get stoped and marked as failed.
    if p_exit_on_fail_bool:
        print("test exited with code - %s"%(p.returncode))
        assert not p.returncode == None # makesure returncode is set by p.wait()
        
        if not p.returncode == 0:
            exit(p.returncode)


# GloFlow application and media management/publishing platform
# Copyright (C) 2020 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

import numpy as np

import tensorflow as tf
import tensorflow.keras.layers
tf.compat.v1.enable_eager_execution() # EAGER_EXEC

sys.path.append("%s/.."%(modd_str))
import gf_ml_data

#----------------------------------------------
def main():
    

    examples_num_int = 10
    epochs_num_int   = 1
    batch_size_int   = 1

    # INPUT_IMG
    # input_img_batch, _ = gf_ml_test_data.download_test_img()
    imgs_train, imgs_test, labels_train, labels_test = gf_ml_data.load__mnist()






    input_l   = tensorflow.keras.layers.Input(shape=(64,))
    dense_1_l = tensorflow.keras.layers.Dense(32, activation="relu", name="layer_1")(input_l)
    dense_2_l = tensorflow.keras.layers.Dense(30, activation="relu", name="layer_2")(dense_1_l)
    dense_3_l = tensorflow.keras.layers.Dense(28, activation="relu", name="layer_3")(dense_2_l)


    dense_softmax_l = tensorflow.keras.layers.Dense(10, activation="softmax", name="layer_4")(dense_3_l)


    


    model = tf.keras.models.Model(inputs=input_l, outputs=dense_softmax_l)
    model.summary()

    # MODEL_COMPILE
    model.compile(optimizer = "adam",
        loss    = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
        metrics = ["accuracy"])

    




    #----------------------------------------------
    def on_batch_end_fun(batch, logs):

        # print("---------------")


        batch_size_int = logs["size"]
        batch_loss_f   = logs["loss"]
        acc_loss_f     = logs["acc"]

        print("loss - %s"%(batch_loss_f))


        # [1:] - exlclude input layer, it does not have weights
        for l in model.layers[1:]:
            # print(type(l))

            # print(l.weights)

            w_var, b_var = l.weights

            print(w_var.shape)

        # print(type(dense_1_l))

        w_var, b_var = model.get_layer("layer_1").weights
        # print(w_var)
        
        # print(w_var.shape)



    #----------------------------------------------
    
    # FIT
    train_tf_history = model.fit(imgs_train[:examples_num_int],
        labels_train[:examples_num_int],
        batch_size      = batch_size_int,
        epochs          = epochs_num_int,
        validation_data = (imgs_test[:examples_num_int], labels_test[:examples_num_int]),
        callbacks       = [tf.keras.callbacks.LambdaCallback(on_batch_end=on_batch_end_fun)])









    # MODEL_PREDICT

    print(imgs_test.shape)
    print(imgs_test[0].shape)

    y = model.predict(imgs_test[:1])

    print(y[0])
    print("pred - %s"%(np.argmax(y[0])))
    print("true - %s"%(labels_test[0]))





#----------------------------------------------
main()


# GloFlow application and media management/publishing platform
# Copyright (C) 2023 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

import pandas as pd

sys.path.append(f"{modd_str}/../gf_core")
import gf_core_sql_db

#---------------------------------------------------------------------------------
# DECORATORS
#---------------------------------------------------------------------------------
def gf_materialize(p_partition_map,
    p_dagster_run_id,
    p_db_client):
    def decorator(p_func):
        def wrapper(*args, **kwargs):
            func_name_str = p_func.__name__
            print(f"gf_materialize decorator ---> {p_func.__name__}")
            
            # CREATE
            materialization_sql_id_int = db_create_materialization(p_partition_map,
                p_dagster_run_id,
                p_db_client)
            
            # FUNC
            try:
                output_map = p_func(*args, **kwargs)
                assert isinstance(output_map, dict)
            except Exception as e:
                
                db_update_materialization("failed", materialization_sql_id_int, p_db_client)
                
                # rethrow the exception further up the chain
                raise
                    
            # COMPLETE
            db_update_materialization("completed", materialization_sql_id_int, p_db_client)

            p_partition_map["output_map"] = output_map 

            return output_map
        return wrapper
    return decorator

#---------------------------------------------------------------------------------
# MAIN
#---------------------------------------------------------------------------------
def get_or_create(p_set_name_str,
    p_group_name_str,
    p_partition_keys_map,
    p_partition_i_int,
    p_partition_size_int,
    p_dagster_asset_id_str,
    p_db_client):
    
    # CHECK_PARTITION_EXISTS
    exists_bool, materialized_bool = db_check_exists(p_set_name_str,
        p_group_name_str,
        p_partition_i_int,
        p_partition_size_int,
        p_db_client)

    
    # CREATE_PARTITION
    if not exists_bool:

        new_partition_map = {
            "set_name_str": p_set_name_str,

            # this is the general group to which this set belongs to
            "group_name_str": p_group_name_str,

            #----------------------
            # PARTITION_KEYS
            "partition_keys_map": p_partition_keys_map,
            
            #----------------------

            "partition_i_int":    p_partition_i_int,
            "partition_size_int": p_partition_size_int,

            "partition_sql_id_int": None,

            # ID of a dagster asset that triggered this partition
            "dagster_asset_id_str": p_dagster_asset_id_str,


            # "output_map": output_map
        }
        
        
        new_partition_sql_id_int = db_create(new_partition_map, p_db_client)
        new_partition_map["partition_sql_id_int"] = new_partition_sql_id_int

        return new_partition_map

    # GET_EXISTING_PARTITION
    else:
        
        # DB
        result = db_get_one(p_set_name_str,
            p_group_name_str,
            p_partition_i_int,
            p_partition_size_int,
            p_db_client)

        partition_sql_id_int, created_at, dagster_asset_id, _, _, _ = result
        _, _, _, materialized_started_at, materialized_actively, materialized = result

        existing_partition_map = {
            "set_name_str":         p_set_name_str,
            "group_name_str":       p_group_name_str,
            "partition_keys_map":   p_partition_keys_map,
            "partition_i_int":      p_partition_i_int,
            "partition_size_int":   p_partition_size_int,
            "partition_sql_id_int": partition_sql_id_int,
            "dagster_asset_id_str": p_dagster_asset_id_str,
        }

        return existing_partition_map

#---------------------------------------------------------------------------------
def db_get_partitions_info(p_db_client):
    
    table_name_str = "gf_data_partitions"
    table_name_materialize_str = "gf_data_partitions_materilize"

    query_str = f"""
        SELECT set_name, group_name, partition_size, ARRAY_AGG(partition_i ORDER BY partition_i) AS partition_ids
        FROM {table_name_str}
        GROUP BY set_name, group_name, partition_size;"""

    cur = p_db_client.cursor()
    cur.execute(query_str)




    rows = cur.fetchall()

    partition_groups_lst = []
    for row in rows:

        set_name_str, group_name_str, partition_size_int, partition_indexes_lst = row
        partition_info_map = {
            "set_name":          set_name_str,
            "group_name":        group_name_str,
            "partition_size":    partition_size_int,
            "partition_indexes": partition_indexes_lst,
            "partitions_num":    len(partition_indexes_lst),
        }

        
        partition_groups_lst.append(partition_info_map)

    df = pd.DataFrame(partition_groups_lst)
    print(df)

    cur.close()

    return df

#---------------------------------------------------------------------------------
def db_create_materialization(p_partition_map,
    p_dagster_run_id_str,
    p_db_client):
    assert isinstance(p_partition_map, dict)

    print("create materialization...")

    table_name_str = "gf_data_partitions_materilize"
    status_str     = "started"


    partition_sql_id_int = p_partition_map["partition_sql_id_int"]
    assert isinstance(partition_sql_id_int, int)

    query_str = f"""
        INSERT INTO {table_name_str} (
            set_name,
            group_name,
            status,
            dagster_run_id,
            partition_id) 
        VALUES (%s, %s, %s, %s, %s)
        RETURNING id;"""

    values_tpl = (p_partition_map["set_name_str"],
        p_partition_map["group_name_str"],
        status_str,
        p_dagster_run_id_str,
        partition_sql_id_int)

    cur = p_db_client.cursor()
    cur.execute(query_str, values_tpl)

    new_sql_id_int = cur.fetchone()[0]


    print(new_sql_id_int)



    p_db_client.commit()
    cur.close()

    return new_sql_id_int

#---------------------------------------------------------------------------------
def db_update_materialization(p_status_str,
    p_materialization_id_int,
    p_db_client):
    assert p_status_str == "failed" or p_status_str == "completed"

    table_name_str = "gf_data_partitions_materilize"
    status_str     = "completed"

    query_str = f"""
        UPDATE {table_name_str}
        SET status = '{p_status_str}'
        WHERE id={p_materialization_id_int};
    """

    cur = p_db_client.cursor()
    cur.execute(query_str)
    p_db_client.commit()
    cur.close()

#---------------------------------------------------------------------------------
def db_create(p_partition_map,
    p_db_client):
    assert isinstance(p_partition_map, dict)

    print("create partition...")

    table_name_str = "gf_data_partitions"

    query_str = f"""
        INSERT INTO {table_name_str} (
            set_name,
            group_name,
            dagster_asset_id,
            partition_i,
            partition_size,
            materialized_actively,
            materialized) 
        VALUES (%s, %s, %s, %s, %s, %s, %s)
        RETURNING id;"""

    values_tpl = (p_partition_map["set_name_str"],
        p_partition_map["group_name_str"],
        p_partition_map["dagster_asset_id_str"],
        p_partition_map["partition_i_int"],
        p_partition_map["partition_size_int"],
        False, # materialized_actively
        True)  # materialized

    cur = p_db_client.cursor()
    cur.execute(query_str, values_tpl)

    new_partition_sql_id_int = cur.fetchone()[0]

    p_db_client.commit()
    cur.close()

    return new_partition_sql_id_int
    
#---------------------------------------------------------------------------------
def db_check_exists(p_set_name_str,
    p_group_name_str,
    p_partition_i_int,
    p_partition_size_int,
    p_db_client):
    
    table_name_str = "gf_data_partitions"
    partition_str = "%s:%s:%s:%s"%(p_set_name_str, p_group_name_str, p_partition_i_int, p_partition_size_int)

    try:
        result = db_get_one(p_set_name_str,
            p_group_name_str,
            p_partition_i_int,
            p_partition_size_int,
            p_db_client)

    except Exception as e:
        if isinstance(e, errors.UndefinedTable):
            # whole table doesnt exist, not just the partition, so report that
            return False
        else:
            # rethrow the exception further up the chain
            raise
    
    if result is not None:
        
        materialized_bool = result[5]

        print(f"partition exists - {partition_str}")
        return True, materialized_bool
    else:
        print(f"partition doesnt exists - {partition_str}")
        return False, False

    cur.close()
    p_db_client.close()

#---------------------------------------------------------------------------------
def db_get_one(p_set_name_str,
    p_group_name_str,
    p_partition_i_int,
    p_partition_size_int,
    p_db_client):

    table_name_str = "gf_data_partitions"
    cur = p_db_client.cursor()
    query_str = f"""
        SELECT
            id,
            created_at,
            dagster_asset_id,
            materialized_started_at,
            materialized_actively,
            materialized
        FROM {table_name_str}
        WHERE set_name = %s AND group_name = %s AND partition_i = %s AND partition_size = %s
    """

    # QUERY
    cur.execute(query_str,
        (p_set_name_str, p_group_name_str, p_partition_i_int, p_partition_size_int))

    result = cur.fetchone()
    return result

#---------------------------------------------------------------------------------
def db_init(p_db_client):
    
    cur = p_db_client.cursor()

    table_name_str = "gf_data_partitions"
    table_partition_materialize_name_str = "gf_data_partitions_materilize"


    #---------------------------------------------------------------------------------
    def create_table():

        # cur.execute(f"DROP TABLE {table_name_str} CASCADE;")
        # cur.execute(f"DROP TABLE {table_partition_materialize_name_str} CASCADE;")

        # GF_DATA_PARTITIONS
        if not gf_core_sql_db.table_exists(table_name_str, cur):

            # dagster_asset_id - rereference to the appropriate Dagster Asset that
            #                    materialized/computed this partition.
            # actively_processed - flag indicating if anyone is activelly processing this 
            # materialized_bool - has this partition been ever materialized?
            create_table_sql_str = f"""
                CREATE TABLE {table_name_str} (
                    id               BIGSERIAL PRIMARY KEY,
                    created_at       TIMESTAMP DEFAULT NOW(),
                    set_name         VARCHAR(255) NOT NULL,
                    group_name       VARCHAR(255) NOT NULL,
                    dagster_asset_id VARCHAR(255) NOT NULL,

                    partition_i    INT NOT NULL,
                    partition_size INT NOT NULL,

                    materialized_started_at TIMESTAMP,
                    materialized_actively   BOOLEAN NOT NULL,
                    materialized            BOOLEAN NOT NULL,

                    UNIQUE(set_name, group_name, partition_i, partition_size)
                );"""
            cur.execute(create_table_sql_str)
            p_db_client.commit()

        # GF_DATA_PARTITIONS_MATERIALIZE
        if not gf_core_sql_db.table_exists(table_partition_materialize_name_str, cur):

            # status         - "started" | "failed" | "completed"
            # dagster_run_id - ID of the Dagster Run within which this materialization is executing
            create_table_sql_str = f"""
                CREATE TABLE {table_partition_materialize_name_str} (
                    id             BIGSERIAL PRIMARY KEY,
                    created_at     TIMESTAMP DEFAULT NOW(),
                    set_name       VARCHAR(255) NOT NULL,
                    group_name     VARCHAR(255) NOT NULL,
                    status         VARCHAR(255) NOT NULL,
                    dagster_run_id VARCHAR(255) NOT NULL,
                    partition_id   BIGINT REFERENCES {table_name_str}(id)
                );"""
            cur.execute(create_table_sql_str)
            p_db_client.commit()
    
    #---------------------------------------------------------------------------------
    create_table()


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import traceback

#--------------------------------------------------------
# p_surrounding_context_attribs_tpl - order matters

def handle_exception(p_exception,
	p_formated_msg_str,
	p_surrounding_context_attribs_tpl,
	p_log_fun):
	assert isinstance(p_exception, Exception)

	if p_formated_msg_str                == None or \
	   p_surrounding_context_attribs_tpl == None:
	
		msg_str = ''
	else:
		assert isinstance(p_surrounding_context_attribs_tpl, tuple)
		msg_str = p_formated_msg_str%(p_surrounding_context_attribs_tpl)
	
	p_log_fun('INFO', 'p_exception.message:%s'%(p_exception.message))
	
	
	with_trace_msg_str = '''
		%s
		exception args:%s
		exception msg :%s
		trace         :%s'''%(msg_str,
			p_exception.args,
			p_exception.message,
			traceback.format_exc())

	p_log_fun('ERROR', with_trace_msg_str)